---
title: Implementing CI/CD around GitLab
author: Lars Wirzenius
documentclass: article
...

# Introduction

![components](component.svg)\ 

This is the plan for the **first iteration** of implementing a CI/CD
system around GitLab for WMF. This iteration tries to do the least
amount of work to prove that the planned architecture is workable.
There are some small differences to the architecture in the planning
document.

See:

* [CI Future WG](https://www.mediawiki.org/wiki/Wikimedia_Release_Engineering_Team/CI_Futures_WG)
* [CI architecure](https://docs.google.com/document/d/1EQuInEV-eY_5kxOZ8E1qEdLr8fb6ihwOD9V_tpVFWuU/edit)

The components are:

* **Gerrit** (or any git server): this is the canonical location for
  the source code. It emits *events* that the controller reacts to.
  Also, the controller sends message to it. This will be mocked in the
  first iteration by using any dumb git server. There will be no
  actual events sent.
* **controller**: this orcestrates builds and deployments; in the
  first iteration, it won't be listening to Gerrit events, and will
  instead have an HTTP API that will be used instead
* **GitLab**: this is used only for it's CI/CD functionality; a
  secondary copy of the git repository is kept here, because GitLab
  requires it
* **Runner**: this is used by GitLab to run builds and tests; it
  uploads built binaries to the artifact store; this corresponds to
  "build worker" in the planning document
* **artifact store**: this stores binaries or other build artifacts so
  they persist when the Runner goes away, given the Runner is a Docker
  container and has no persistency
* **VCS worker**: this retrieves source code from Gerrit (or other git
  server) and pushes is to GitLab; it's a separate system so it can be
  given credentials to access non-public git repositories
* **deployment worker**: this gets binaries from the artifact store
  and deploys them to the test environment
* **test environment**: this mocks a production-like environment for
  running sites and services; for the first prototype, it just
  publishes the binaries

Differences from the planning document:

* There's GitLab to run builds, rather then the controller commanding
  build workers directly.
* There's no log store. This isn't necessary for the first iteration.
* There's only one environment, the test environment, and it won't be
  running sites or services. Deployment is simulated by merely
  publishing the build artifacts in the test environment.

# New components

There will need to be some new components. We'll keep them as simple
as possible. Most will have a simple HTTP API. We'll use signed JWT
access tokens, which will be generated staticlly and installed when
the components are set up. (This is highly inadequtate for production,
but this is just the first iteration.)

All HTTP APIs will served over HTTPS using a TLS certificate from
Let's Encrypt, for the first iteration.

None of the components will aim to be fast or to serve many clients,
in the first iteration. They'll be implemented using haproxy (TLS),
bottle.py, and some custom Python code. All of this may change after
the first iteration, but these tools are familiar to me so I can just
use them, and don't have to learn stuff to get started.

## Controller

* Simple HTTP API
* Endpoint: POST /cd, body specifies which repo and ref to build and
  deploy; queues the build, the queued build will be visible via the
  /status endpoint

        {
            "repo": "...",
            "ref": "..."
            "artifact": "..."
        }

* Endpoint: GET /status, which lists what jobs (posts to /cd) are
  queued, or running, or finished

        {
            "builds": {
                "name": "status"
            }
        }

## VCS worker

* Simple HTTP API
* Endpoint: POST /repo, body specifies repo and ref to fetch; git
  clone (or pulls) that, and pushes to GitLab repo with same name
  (creates it if necessary), but using only the master branch; returns
  info of how things went

        {
            'git': 'git://git.liw.fi/heippa',
            'ref': 'master',
            'name': 'hithere2'
        }

* Clones the specified git repository, and pushes the specified branch
  as master in GitLab. Deletes the repository on GitLab if it exists
  already (for hygiene). If there's a `.gitlab-ci.yml` in the repo,
  GitLab CI will run it.

## Artifact store

* Simple HTTP API
* For the prototype, this is the Ick artifact store for this
* Endpoint: PUT /blobs/NAME, stores body as blob named NAME,
  overwriting it if it existed already
* Endpoing: GET /blobs/NAME, returns blob, with a generic content
  type, or 404 if not found

## Deployment worker

* Simple HTTP API
* Endpoint: POST /deploy, body specifies name in artifact store,
  copies that to the test environment using SSH

        {
            "artifact_id": "foo",
            "published_name": "foo-1.2.4"
        }


## Test environment

* SSH server with HTTP server serving a directory
* Deployment is mocked by copying blob from artifact store to the
  directory being serveed by HTTP server

# How a build works

![Build sequence](buildseq.svg)\ 


The build has the following steps:

*  Gerrit notifies the controller of a change in a git repo.

*  The controller tells the VCS worker to copy the repo to GitLab, by
   doing a "POST /repo" HTTP request.

*  The VCS worker git clones the repo from Gerrit. It may do a git
   pull instead to update an existing clone, but that's an
   optimisation, whiche we'll implement when it's needed. The git
   operation may require credentials (e.g., security embargoed
   repositories), which the VCSWorker has: they're installed when the
   host is deployed. No other CI host has those credentials.

*  The VCS worker pushed the repository to GitLab. This may again
   require credentials. The push triggers GitLab CI to run the commit
   stage build and test command specified in `.gitlab-ci.yml`.

*  The VCS worker responds to the POST request from the controller
   with results.

*  GitLab tells the Runner host to run the build and test commands.
   The runner does that.

*  The Runner uploads any binaries it builds to the artifact store.

*  The Runner tells GitLab it's finished.

*  GitLab tells the controller via a webhook that a build has
   finished.

*  The controller tells the deployer to start a deployment, via a
   "POST /deploy" HTTP API call.

*  The deployer fetches the artifacts it has been told to deploy
   from the artifact store.

*  The deployer copyies the artifacts to the test environment. (In a
   future iteration this will be a more sophisticated deployment
   process.)

*  The deployer responds to the HTTP request from the controller with
   the results.

*  The controller notifies Gerrit of a build and deployment having
   been finished. (Except not in the first iteration.)

# Testing the CI system

We don't test invidual components here, only the CI system as a whole.
The test consists of building a small C program of the "hello, world"
variety, and checking that the result works. The repository is at
<http://git.liw.fi/heippa> and contains a `.gitlab-ci.yml` file, which
GitLab automatically uses to build the project. The assisting services
around GitLab then publish the resulting binary to a web server, as
outlineed above.

[Fable]: https://fable.liw.fi/

To run this test suite, install [Fable][] (see tutorial for
installation instructions) and then run this command in the source
tree of this test suite:

```sh
WMF_TOKEN="(create-token key iss aud 'trigger status')" ftt-codegen --run gitlab.md
```

First, we trigger the build by telling the controller that a
repository has changed.

```fable
given a git repository git://git.liw.fi/heippa with a simple C program
when I trigger a build with {"git": "git://git.liw.fi/heippa", "ref": "master", "gitlab": "hithere"}
then HTTP status code is 200
```

The controller keeps track of running builds, although it can take a
few seconds for the build status to change. So we wait a bit and then
check. (This is a simple, but not very good way of implementing this.)

```fable
when I wait for 5 seconds
and I check build status
then response body indicates hithere is building
```

Then we check that the build succeeds. Again, the build may take a
while, so we wait a minute.

```fable
when I wait for 60 seconds
and I check build status
then response body indicates hithere is success
```

Finally, we retrieve the published binary and run it.

```fable
when I retrieve http://wmf2-testenv.vm.liw.fi/hithere to a local file hithere
and I run ./hithere
then the output is "hello, world"
```

Fin.
